--[[ the main WE2 window that contains the builders, fast builders, recovery skills, the default gambit skill, and the menu button

For purposes of this plugin:
Builders are the main warden building skills (normally spear = 1, shield = 2, fist = 3)
Fast builders are the double building skills (f/f, sp/sp, sh/sh, and the other combos of f/sp/sh)
Recovery skills refer to the Recovery and Quick Recovery skills
Default gambit refers to the default gambit skill that changes to the actual gambit skill when an active gambit can be used
Gambits refer to the 2-5 sequence of builder skills
]]--

import "Turbine.Gameplay";
import "Turbine.UI.Lotro";
import "SDRPlugins.WardenEase2.WEQuickslotWindow";
import "SDRPlugins.WardenEase2.VindarPatch"; -- fixes problems with European clients

WEController = SDRPlugins.WardenEase2.class( Turbine.UI.Window );

function WEController:Constructor()
	Turbine.UI.Window.Constructor( self );

	player = Turbine.Gameplay.LocalPlayer.GetInstance();

	self.allowedChange = 0; -- prevents quickslots from being changed
	self.spacing = 2;
	self.extraSpacing = 3;
	self.quickslotSize = 36; -- the normal quickslot size (or the size of the pictures) is 36 plus 2 on all sides
	self.quickslotsBackColors = {}; -- back colors for builder/gambit quickslots
	self.builderKeyOverlay = {}; -- overlays for builder/gambit quickslots
	self.quickslots = { }; -- contains all quickslots
	self.builder = { }; -- contains all buidlers
	self.gambitBar = { }; -- contains the bars that hold the sp/fi/sh gambits

	self:LoadSettings();
	self:SetPosition( self.settings.left, self.settings.top );
	self:SetBackColor(Turbine.UI.Color(0, 0, 0, 0));

	levelChangedHandler1 = function( sender, args )
		self:RefreshQuickslots();
	end

	AddCallback(player, "LevelChanged", levelChangedHandler1);
	
	-- creates the menu button on the main WE window
	self.menuButton = Turbine.UI.Control();
	self.menuButton:SetParent( self );
	self.menuButton:SetBackground( "SDRPlugins/WardenEase2/Resources/DownArrow.tga" );
	self.menuButton:SetSize( 14, 13 );
	
	-- when the user clicks the down arrow, this action fires
	self.menuButton.MouseDown = function( sender, args )
		local contextMenu = Turbine.UI.ContextMenu()
		local contextMenuItems = contextMenu:GetItems();
		--local barType = { "Spear", "Fist", "Shield" };

		-- creates the individual submenu parts
		for k,v in pairs( { "Spear", "Fist", "Shield" } ) do

			-- labels the submenu part based on Spear/First/Shield
			local subMenu = Turbine.UI.MenuItem( SDRPlugins.WardenEase2[v] );

			-- set/don't set based on settings
			subMenu:SetChecked( self.settings.enabled[v] );

			-- handler for when the user selects one of the submenu parts
			subMenu.Click = function( sender, args )

				-- switch the enabled/disabled of the bar clicked
				self.settings.enabled[v] = not self.settings.enabled[v];

				-- show a bar that has been enabled
				if self.settings.enabled[v] then
					self.gambitBar[v] = SDRPlugins.WardenEase2.WEQuickslotWindow(v, self.settings[v]);

				-- hide a bar that has been disabled
				else
					self.gambitBar[v]:SetVisible(false);
					self.gambitBar[v]:SetParent(nil);
					self.gambitBar[v] = nil;
				end
				
				-- save settings
				self:SaveSettings();
			end

			-- adds the context sub menu items to the menu
			contextMenuItems:Add( subMenu );
		end


		-- the next section allows the user to turn on/off descriptions
		local subMenu;
		if self.settings.Spear.descriptionsOn then
			subMenu = Turbine.UI.MenuItem( SDRPlugins.WardenEase2.hideDescriptions );
		else
			subMenu = Turbine.UI.MenuItem( SDRPlugins.WardenEase2.showDescriptions );
		end
		subMenu.Click = function( sender, args )
			for k,v in pairs( { "Spear", "Fist", "Shield" } ) do
				-- set the descriptions on/off whether the bar is showing to keep them all showing/not showing
				self.settings[v].descriptionsOn = not self.settings[v].descriptionsOn;

				-- if the bar is nil it is hidden, do not bother destroying and recreating the bar
				if self.gambitBar[v] ~= nil then
					local bDraggable = self.gambitBar[v].DragBar:IsDraggable();

					self.gambitBar[v]:SetVisible(false);
					self.gambitBar[v]:SetParent(nil);
					self.gambitBar[v].DragBar = nil;
					self.gambitBar[v] = nil;
					self.gambitBar[v] = SDRPlugins.WardenEase2.WEQuickslotWindow( v, self.settings[v] );

					if bDraggable then
						self.gambitBar[v].DragBar:ToggleDraggable();
					end
				end
			end
			self:SaveSettings();
		end
		-- adds the submenu item
		contextMenuItems:Add( subMenu );

		local subMenu = Turbine.UI.MenuItem("---------------------------------------");
		contextMenuItems:Add(subMenu);
		
		-- adds the key overlay options to the menu
		local twoIndex = 6;
		local foursAdded = 0;
		for k, v in pairs(self.settings.builderKeyOverlay) do
			if k ~= "show" then
				local subMenu = Turbine.UI.MenuItem(v.." - "..SDRPlugins.WardenEase2[k]);
				subMenu.Click = function(sender, args)
					self.settings.builderKeyOverlay[k] = v + 1;
					if self.settings.builderKeyOverlay[k] > 4 then
						self.settings.builderKeyOverlay[k] = 1;
					end
					self:SaveSettings();
					self:RefreshLayout();
				end

				-- insert the menu items in numerical order
				if v == 1 then
					contextMenuItems:Insert(6, subMenu);
					twoIndex = twoIndex + 1;
				elseif v == 4 then
					contextMenuItems:Add(subMenu);
					foursAdded = foursAdded + 1;
				elseif v == 2 then
					contextMenuItems:Insert(twoIndex, subMenu);
				elseif v == 3 then
					contextMenuItems:Insert((contextMenuItems:GetCount() - foursAdded)+1, subMenu);
				end
			end
		end

		local subMenu = Turbine.UI.MenuItem("---------------------------------------");
		contextMenuItems:Add(subMenu);

		local subMenu;
		if self.settings.builderKeyOverlay["show"] then
			subMenu = Turbine.UI.MenuItem(SDRPlugins.WardenEase2.hideKeyOverlay);
		else
			subMenu = Turbine.UI.MenuItem(SDRPlugins.WardenEase2.showKeyOverlay);
		end
		subMenu.Click = function(sender, args)
			self.settings.builderKeyOverlay["show"] = not self.settings.builderKeyOverlay["show"];
			self:SaveSettings();
			self:RefreshLayout();
		end
		contextMenuItems:Add(subMenu);

		local subMenu;
		subMenu = Turbine.UI.MenuItem(SDRPlugins.WardenEase2.reverseWE2Window);
		subMenu.Click = function(sender, args)
			if self.settings.reversed == 0 then
				self.settings.reversed = 1;
			else self.settings.reversed = 0;
			end
			self:SaveSettings();
			self:RefreshLayout();
		end
		contextMenuItems:Add(subMenu);

		local subMenu;
		if self.settings.fastBuilders["show"] then
			subMenu = Turbine.UI.MenuItem(SDRPlugins.WardenEase2.hideFastBuilders);
		else
			subMenu = Turbine.UI.MenuItem(SDRPlugins.WardenEase2.showFastBuilders);
		end
		subMenu.Click = function(sender, args)
			self.settings.fastBuilders["show"] = not self.settings.fastBuilders["show"];
			self:SaveSettings();
			self:RefreshQuickslots();
		end
		contextMenuItems:Add(subMenu);

		local subMenu;
		if self.settings.recoverySkills["show"] then
			subMenu = Turbine.UI.MenuItem(SDRPlugins.WardenEase2.hideRecoverySkills);
		else
			subMenu = Turbine.UI.MenuItem(SDRPlugins.WardenEase2.showRecoverySkills);
		end
		subMenu.Click = function(sender, args)
			self.settings.recoverySkills["show"] = not self.settings.recoverySkills["show"];
			self:SaveSettings();
			self:RefreshQuickslots();
		end
		contextMenuItems:Add(subMenu);

		-- shows the context menu when the down arrow is clicked
		contextMenu:ShowMenu();
	end
	-- end down arrow mouse click handler

	-- when the position of the window has changed, update settings
	self.PositionChanged = function( sender, args )
		self.settings.left, self.settings.top = self:GetPosition();
		self:SaveSettings();
	end

	self:RefreshQuickslots();
	self.DragBar = SDRPlugins.WardenEase2.UI.DragBar( self, SDRPlugins.WardenEase2.we2MainWindow );

	-- goes through the settings.enabled table
	-- k is spear/fist/shield, and v is true/false
	for k,v in pairs(self.settings.enabled) do
		if v then
			-- sets the window to show if the window should show (duh!)
			self.gambitBar[k] = SDRPlugins.WardenEase2.WEQuickslotWindow(k, self.settings[k]);
		end
	end
	
	-- sets the main WE window to show (which should almost always be true)
	self:SetVisible( true );
	-- opacity should be full
	self:SetOpacity( 1 );

	maxGambitCountChangedHandler = function(sender, args)
		self:RefreshQuickslots();
	end

	AddCallback(Turbine.Gameplay.Attributes.WardenAttributes, "MaxGambitCountChanged", maxGambitCountChangedHandler);
end

--Fix broken settings (due to corrupted data or pre-v1.1 settings)
function PLoadSettings()
	return pcall(function () PatchDataLoad( Turbine.DataScope.Character, "WardenEase") end)
end

-- loads the user settings
function WEController:LoadSettings()

	euroFormat=(tonumber("1,000")==1); -- will be true if the number is formatted with a comma for decimal place, false otherwise
	-- create a function for automatically converting a number in string format to its correct numeric value
	if euroFormat then
 		function euroNormalize(value)
   			return tonumber((string.gsub(value,"%.",",")));
		end
	else
   		function euroNormalize(value)
   			return tonumber((string.gsub(value,",",".")));
  		end
	end

	if PLoadSettings() then
		self.settings = PatchDataLoad( Turbine.DataScope.Character, "WardenEase");
	end

	-- make sure the settings are a table, if not, create new settings
	if type( self.settings ) ~= "table" then
		self.settings = { };
	end

	-- for people coming from previous versions not updated by me, we need to blank their data since it can be corrupted
	if self.settings.version == nil then
		self.settings = {};
	end

	self.settings.version = version;
	self.settings.version = euroNormalize(self.settings.version); -- European fix

	-- allows the user to show key overlays ove the builder/gambit buttons
	if type(self.settings.builderKeyOverlay) ~= "table" then
		self.settings.builderKeyOverlay = {};
		self.settings.builderKeyOverlay["show"] = false;
		self.settings.builderKeyOverlay["Spear"] = 1;
		self.settings.builderKeyOverlay["Shield"] = 2;
		self.settings.builderKeyOverlay["Fist"] = 3;
		self.settings.builderKeyOverlay["Gambit"] = 4;
	end
	-- Europe client fixes
	self.settings.builderKeyOverlay["Spear"] = euroNormalize(self.settings.builderKeyOverlay["Spear"]);
	self.settings.builderKeyOverlay["Shield"] = euroNormalize(self.settings.builderKeyOverlay["Shield"]);
	self.settings.builderKeyOverlay["Fist"] = euroNormalize(self.settings.builderKeyOverlay["Fist"]);
	self.settings.builderKeyOverlay["Gambit"] = euroNormalize(self.settings.builderKeyOverlay["Gambit"]);

	-- allows the user to show the fast builder skills
	if type(self.settings.fastBuilders) ~= "table" then
		self.settings.fastBuilders = {};
		self.settings.fastBuilders["show"] = true;
	end

	-- allows the user to show the recovery skills
	if type(self.settings.recoverySkills) ~= "table" then
		self.settings.recoverySkills = {};
		self.settings.recoverySkills["show"] = true;
	end

	-- make sure that the settings.enabled is a table, otherwise set defaults
	if type( self.settings.enabled ) ~= "table" then
		self.settings.enabled = { };
		self.settings.enabled["Spear"] = true;
		self.settings.enabled["Shield"] = true;
		self.settings.enabled["Fist"] = true;
	end

	-- default fast builder locations
	if type( self.settings.fastBuilderLocs ) ~= "table" then
		self.settings.fastBuilderLocs = { };
		self.settings.fastBuilderLocs[1] = "Spear and Spear";
		self.settings.fastBuilderLocs[2] = "Shield and Shield";
		self.settings.fastBuilderLocs[3] = "Fist and Fist";
		self.settings.fastBuilderLocs[4] = "Spear and Shield";
		self.settings.fastBuilderLocs[5] = "Shield and Spear";
		self.settings.fastBuilderLocs[6] = "Fist and Spear";
		self.settings.fastBuilderLocs[7] = "Spear and Fist";
		self.settings.fastBuilderLocs[8] = "Shield and Fist";
		self.settings.fastBuilderLocs[9] = "Fist and Shield";
	end

	-- determines whether the controller window is built up or down
	if self.settings.reversed == nil then
		self.settings.reversed = 0;
	end
	self.settings.reversed = euroNormalize(self.settings.reversed); -- European fix

	-- users display width and height
	local displayWidth = Turbine.UI.Display:GetWidth();
	local displayHeight = Turbine.UI.Display:GetHeight();

	-- sets window to center of screen if settings.left are nil
	-- sets window to center of screen if the settings.left are outside the screen area
	if self.settings.left == nil or self.settings.left > displayWidth or self.settings.left < 0 then
		self.settings.left = Turbine.UI.Display:GetWidth() / 2 - 7 * (self.quickslotSize + self.spacing) / 2;
	end
	self.settings.left = euroNormalize(self.settings.left); -- European fix

	-- sets window to center of screen if settings.top are nil
	-- sets window to center of screen if settings.top are outside the screen area
	if self.settings.top == nil or self.settings.top > displayHeight or self.settings.top < 0 then
		self.settings.top = Turbine.UI.Display:GetHeight() * 3 / 4;
	end
	self.settings.top = euroNormalize(self.settings.top); -- European fix

	-- initializes the bars for Spear/Fist/Shield
	for k,v in pairs({ "Spear", "Fist", "Shield" }) do
		-- makes sure each setting.Fist/Shield/Spear is a table
		if type( self.settings[v] ) ~= "table" then
			self.settings[v] = { };
		end

		-- if the settings[Fist/Shield/Spear] is nil or outside of screen area, then create defaults
		if self.settings[v].left == nil or self.settings[v].left > displayWidth or self.settings[v].left < 0 then
			--[[ 
			create each bar slightly off from each other so they do not overlap
			basic idea is first bar is ScreenWidth * 2/3, while
			second bar is ScreenWidth * 2/3 + self.spacing + 32 (where self.spacing is 2 and 32 is the size of the pictures), and
			third bar is ScreenWidth * 2/3 + (self.spacing + 32) * 2
			]]--
			self.settings[v].left = Turbine.UI.Display:GetWidth() * 2 / 3;
			if v == "Fist" then
				self.settings[v].left = self.settings[v].left + self.spacing + 32;
			end
			if v == "Shield" then
				self.settings[v].left = self.settings[v].left + 2 * (self.spacing + 32);
			end
		end
		self.settings[v].left = euroNormalize(self.settings[v].left); -- European fix

		-- if the settings[Fist/Shield/Spear] is nil or outside of screen area, then create defaults
		if self.settings[v].top == nil or self.settings[v].top > displayHeight or self.settings[v].top < 0 then
			--[[
			create each bar slightly off from each other (because of differences in number of gambits - this might need to be changed (or not since it's just defaults))
			basic idea is first bar is ScreenHeight / 3, while
			second bar is ScreenHeight / 3 + self.spacing + 32 (where self.spacing is 2 and 32 is the size of the pictures), and
			third bar is ScreenHeight / 3 + (self.spacing + 32) * 2
			]]--
			self.settings[v].top = Turbine.UI.Display:GetHeight() / 3;
			if v == "Fist" then
				self.settings[v].top = self.settings[v].top + self.spacing + 32;
			end
			if v == "Shield" then
				self.settings[v].top = self.settings[v].top + 2 * (self.spacing + 32);
			end
		end
		self.settings[v].top = euroNormalize(self.settings[v].top); -- European fix

		-- defaults if the bar should be vertical or horizontal (this value is false if it is vertical)
		if type( self.settings[v].transposed ) ~= "boolean" then
			self.settings[v].transposed = false;
		end

		-- default if descriptions should be shown
		if type( self.settings[v].descriptionsOn ) ~= "boolean" then
			self.settings[v].descriptionsOn = false;
		end

		-- default hidden gambits on each bar
		if type( self.settings[v].hiddenGambits ) ~= "table" then
			self.settings[v].hiddenGambits = {};
		end
	end
end
-- end of loading settings

-- saves settings
function WEController:SaveSettings()
	pcall(function() PatchDataSave( Turbine.DataScope.Character, "WardenEase", self.settings) end);
end

--[[
Sets the gambit that is displayed in the main WE window
skillInfo is the Turbine.Gameplay.SkillInfo object of the Gambit that you want to set
]]--
function WEController:SetGambit( skillInfo )

	self.builder = {};
	
	--self.gambitLength = skillInfo:GetGambitCount(); -- use once API is fixed

	-- Post HD fix
	local GambitTable = SDRPlugins.WardenEase2.Gambits;
	self.gambitLength = #GambitTable[skillInfo:GetName()];
	self.maxGambitLength = player:GetClassAttributes():GetMaxGambitCount();
	-- end HD fix
	
	-- for the builder slots that won't be needed, set them to nil
	for i = 1, self.maxGambitLength - self.gambitLength do
		self.builder[i] = nil;
	end

	-- for the rest, set the builder that needs to be there
	for i = self.maxGambitLength - self.gambitLength + 1, self.maxGambitLength do
		--self.builder[i] = skillInfo:GetGambit(i + self.gambitLength - self.maxGambitLength); -- use once API is fixed
		-- HD Fix
		self.builder[i] = GambitTable[skillInfo:GetName()][i + self.gambitLength - self.maxGambitLength];
		-- end HD fix
	end
	self:RefreshQuickslots();
end

-- refreshes quickslots
function WEController:RefreshQuickslots()
	local i, k, v;
	self.maxGambitLength = player:GetClassAttributes():GetMaxGambitCount();
	self.nRecoverySkills = 0;
	self.nFastBuildSkills = 0;

	--[[
	Number of skills that need quickslots
	Starts at number of gambit slots plus 1
	This makes sure the number of gambit slots have quickslots and the default skill has a quickslot
	]]--
	local skills = self.maxGambitLength + 1;

	-- create tables for the different skills that we have
	self.recoverySkills = {[SDRPlugins.WardenEase2.QuickRecovery]=false, [SDRPlugins.WardenEase2.Recovery]=false};

	self.builderSkills = {[SDRPlugins.WardenEase2.Fist]=false, [SDRPlugins.WardenEase2.Shield]=false, [SDRPlugins.WardenEase2.Spear]=false};

	self.trainedSkills = player:GetTrainedSkills();

	for i = 1, self.trainedSkills:GetCount() do
		local itemIName = self.trainedSkills:GetItem(i):GetSkillInfo():GetName();
		
		-- sorts known skills based on what we need
		-- if the skill name contains recovery, it's one of the recovery skills, set it to true
		if itemIName == SDRPlugins.WardenEase2.Recovery or itemIName == SDRPlugins.WardenEase2.QuickRecovery then
			self.recoverySkills[itemIName] = true;
			skills = skills + 1;
			self.nRecoverySkills = self.nRecoverySkills + 1;
		
		-- if the skill name contains quick, it is the spear builder, set it to true
		elseif itemIName == SDRPlugins.WardenEase2.Spear then
			self.builderSkills[SDRPlugins.WardenEase2.Spear] = true;
		-- if the skill name contains shield
		elseif itemIName == SDRPlugins.WardenEase2.Shield then
			self.builderSkills[SDRPlugins.WardenEase2.Shield] = true;
		-- if the skill name contains Warden's, it is the fist builder, set it to true
		elseif itemIName == SDRPlugins.WardenEase2.Fist then
			self.builderSkills[SDRPlugins.WardenEase2.Fist] = true;
		-- we don't have to check for alternative's of these names because the skill always returns the default name
		
		-- if the word "and" is in the skill name, it might be a fast builder skill
		elseif string.find(itemIName, SDRPlugins.WardenEase2.andWord) ~= nil then
			--[[ the following if statements take the basic following format:
				if the skill contains "BUILDER_NAME and" then it is a fast builder
			]]--
			if string.find(itemIName, SDRPlugins.WardenEase2.FistWord .. " " .. SDRPlugins.WardenEase2.andWord) ~= nil then
				skills = skills + 1;
				self.nFastBuildSkills = self.nFastBuildSkills + 1;
			elseif string.find(itemIName, SDRPlugins.WardenEase2.ShieldWord .. " " .. SDRPlugins.WardenEase2.andWord) ~= nil then
				skills = skills + 1;
				self.nFastBuildSkills = self.nFastBuildSkills + 1;
			elseif string.find(itemIName, SDRPlugins.WardenEase2.SpearWord .. " " .. SDRPlugins.WardenEase2.andWord) ~= nil then
				skills = skills + 1;
				self.nFastBuildSkills = self.nFastBuildSkills + 1;
			end
			-- again we don't have to check for alternatives to fast builder skills (like "Fist and Javelin") because the name always returns the default name
		end
	end

	if(not self.settings.fastBuilders["show"]) then
		skills = skills - self.nFastBuildSkills;
		self.nFastBuildSkills = 0;
	end
	if(not self.settings.recoverySkills["show"]) then
		skills = skills - self.nRecoverySkills;
		self.nRecoverySkills = 0;
	end

	--[[
	Remove any quickslots no longer required, which should really only happen if a Warden loses a skill.
	Basically, if the number of skills in the table self.skills+1 is less than the number of quickslots, for each skill greater than the quickslots do
	]]--
	for i = (skills + 1), #(self.quickslots), 1 do

		-- remove visibility of the item
		-- always remove #self.skills+1 because we'll be deleting it (thus a new one is in the place next time around - do NOT use i
		self.quickslots[(skills + 1)]:SetVisible( false );
		-- remove the parent
		self.quickslots[(skills + 1)]:SetParent( nil );
		-- delete from table
		table.remove(self.quickslots, (skills + 1));
	end
	
	-- Add any new quickslots that should be created. Creates the actual quickslots for each skill needing a quickslot
	for i = #(self.quickslots) + 1, skills, 1 do
		self.quickslots[i] = Turbine.UI.Lotro.Quickslot();
		self.quickslots[i]:SetParent( self );
		self.quickslots[i]:SetSize( self.quickslotSize, self.quickslotSize );
		self.quickslots[i]:SetAllowDrop( false ); -- prevents quickslot from being overwritten

		-- in the off chance a user changes a shortcut without permission, refresh the layout to reset quickslots
		self.quickslots[i].ShortcutChanged = function( sender, args )
			if self.allowedChange == 0 then
				self:RefreshLayout();
			end
		end
	end

	self:RefreshLayout();
end

-- debugging purposes
function WEController:pvar(tmp)
	Turbine.Shell.WriteLine(tmp);
end

-- refreshes the layout
function WEController:RefreshLayout()
	local displayWidth = Turbine.UI.Display:GetWidth();
	local displayHeight = Turbine.UI.Display:GetHeight();
	local modifier = 1;

	self.allowedChange = self.allowedChange + 1; -- allow changes to be made

	-- sets the width to the max number of gambits and recovery skills + 1 (which is for the default gambit skill)
	self.slotsWide = self.maxGambitLength + self.nRecoverySkills + 1;

	-- if there's fast building skills, then there needs to be 3 slots high (since I place the fast builders in a 3x3 grid)
	if self.nFastBuildSkills > 6 then
		self.slotsHigh = 4;
	elseif self.nFastBuildSkills > 3 then
		self.slotsHigh = 3;
	elseif self.nFastBuildSkills > 0 then
		self.slotsHigh = 2;
	else
		self.slotsHigh = 1;
	end

	local i, width, height;

	-- determines width of the window based on different factors
	-- takes number of slots needed (slotWide) and times it by the slot spacing for each quickslot (self.spacing + self.quickslotSize) and then subtracts the extra spacing at the end
	width = self.slotsWide * (self.spacing + self.quickslotSize) - self.spacing;
	
	-- if there are recovery skills, the width needs to be wider (to place a bit of space between the default gambit skill and the recovery skills)
	if self.nRecoverySkills > 0 then
		width = width + self.extraSpacing;
	end

	-- determines the height of the window based on certain factors
	height = (self.slotsHigh - 1) * (self.spacing + self.extraSpacing) + self.slotsHigh * self.quickslotSize + self.spacing + self.menuButton:GetHeight();

	self:SetSize(width, height);
	local rHeightLoc = self.settings.reversed * (height - self.quickslotSize); -- - self.quickslotSize); -- when the loc element is subtracted from this number it will switch the layout of the main window
	local multiplier = -1; -- used to correct negative numbers during normal orientation
	if rHeightLoc ~= 0 then
		multiplier = 1;
	end

	for i = 1, self.maxGambitLength+1, 1 do
		-- create the control behind the quickslot that will change color based on correct sequencing of gambits
		if self.quickslotsBackColors[i] == nil then -- only create stuff if it is nil, eliminates memory leak
			self.quickslotsBackColors[i] = Turbine.UI.Control();
			self.quickslotsBackColors[i]:SetSize( self.quickslotSize + self.spacing, self.quickslotSize + self.spacing);
			self.quickslotsBackColors[i]:SetParent( self);
			self.quickslotsBackColors[i]:SetVisible(true);
			self.quickslotsBackColors[i]:SetAllowDrop( false );
			self.quickslotsBackColors[i]:SetZOrder(-100);
		end
		self.quickslotsBackColors[i]:SetPosition( (i - 1) * (self.quickslotSize + self.spacing), (rHeightLoc - 0) * multiplier );
	end
	-- 0 if the gambit sequence is wrong at some point
	-- 1 if the gambit sequence isn't complete yet (but is still correct up to that point)
	-- 2 if the gambit sequence is complete (and anything after is ignored)
	local completeGambit = 1;

	if self.settings.builderKeyOverlay["show"] then
		for i = 1, self.maxGambitLength+1, 1 do
			-- create the control for the key overlays
			if self.builderKeyOverlay[i] == nil then -- only create stuff if it is nil, eliminates memory leak
				self.builderKeyOverlay[i] = Turbine.UI.TextBox();
				self.builderKeyOverlay[i]:SetSize( self.quickslotSize, self.quickslotSize );
				self.builderKeyOverlay[i]:SetParent( self);
				self.builderKeyOverlay[i]:SetAllowDrop( false );
				self.builderKeyOverlay[i]:SetZOrder(250);
				self.builderKeyOverlay[i]:SetFont(Turbine.UI.Lotro.Font.TrajanProBold36);
				self.builderKeyOverlay[i]:SetFontStyle(Turbine.UI.FontStyle.Outline);
				self.builderKeyOverlay[i]:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleCenter);
				self.builderKeyOverlay[i]:SetMouseVisible(false);
				self.builderKeyOverlay[i]:SetReadOnly(true);
				self.builderKeyOverlay[i]:SetSelectable(false);
				self.builderKeyOverlay[i]:SetVisible(true);
			end
			self.builderKeyOverlay[i]:SetPosition( (i - 1) * (self.quickslotSize + self.spacing), (rHeightLoc - 0) * multiplier );
			if i == self.maxGambitLength+1 then
				self.builderKeyOverlay[i]:SetText(self.settings.builderKeyOverlay["Gambit"]);
			elseif self.builder[i] == 1 or self.builder[i] == 4 then -- spear or javelin, place the Spear overlay
				self.builderKeyOverlay[i]:SetText(self.settings.builderKeyOverlay["Spear"]);
			elseif self.builder[i] == 2 then -- shield
				self.builderKeyOverlay[i]:SetText(self.settings.builderKeyOverlay["Shield"]);
			elseif self.builder[i] == 3 then -- fist
				self.builderKeyOverlay[i]:SetText(self.settings.builderKeyOverlay["Fist"]);
			else self.builderKeyOverlay[i]:SetText(nil);
			end
		end
	else
		for i = 1, self.maxGambitLength+1, 1 do
			if self.builderKeyOverlay[i] ~= nil then -- only hide controls that are not nil (to prevent errors)
				self.builderKeyOverlay[i]:SetText(nil);
			end
		end
	end

	if self.gambitLength == nil then
		self.gambitLength = 0;
	end

	-- sets the number of blank building slots for the user
	local gambitSequence = self.gambitLength - player:GetClassAttributes():GetGambitCount();
	for i = 1, self.maxGambitLength, 1 do
		self.quickslots[i]:SetShortcut( Turbine.UI.Lotro.Shortcut( Turbine.UI.Lotro.ShortcutType.Undefined, "" ) ); -- resets quickslot so that the tooltips are right
		if self.builder[i] ~= nil then -- self.builder[i] refers to the builder needed for the gambit
			if self.builder[i] == 1 or self.builder[i] == 4 then -- spear or javelin, place the Spear skill
				self.quickslots[i]:SetShortcut( Turbine.UI.Lotro.Shortcut( Turbine.UI.Lotro.ShortcutType.Skill, SDRPlugins.WardenEase2.Skills.Spear.Data ) );
			elseif self.builder[i] == 2 then -- shield
				self.quickslots[i]:SetShortcut( Turbine.UI.Lotro.Shortcut( Turbine.UI.Lotro.ShortcutType.Skill, SDRPlugins.WardenEase2.Skills.Shield.Data ) );
			elseif self.builder[i] == 3 then -- fist
				self.quickslots[i]:SetShortcut( Turbine.UI.Lotro.Shortcut( Turbine.UI.Lotro.ShortcutType.Skill, SDRPlugins.WardenEase2.Skills.Fist.Data ) );
			end
		end
		-- sets the position of the quickslot based on the number of builders put in already
		self.quickslots[i]:SetPosition( (i - 1) * (self.quickslotSize + self.spacing), (rHeightLoc - 0) * multiplier );

		-- if there are more gambits than the set gambit length, check sequence
		if gambitSequence < 0 then
			-- multiply the gambitSequence by -1 to get where the beginning gambit should be
			local oppositeSequence = gambitSequence * -1;
			
			-- if that is less than 0, the gambit slots are empty, set them to black
			if i+(player:GetClassAttributes():GetGambitCount() - self.maxGambitLength) <= 0 then
				self.quickslotsBackColors[i]:SetBackColor(Turbine.UI.Color( 0,0,0 ));
			-- check to make sure oppositeSequence+i and i+(gambitCount-maxGambitLength) are within correct ranges
			elseif oppositeSequence+i <= self.maxGambitLength and i+(player:GetClassAttributes():GetGambitCount() - self.maxGambitLength) <= player:GetClassAttributes():GetGambitCount() then
				-- if the gambits are the same, and the gambit isn't wrong, then color the spot green
				-- gets the builder at the position, changes it to 3 if it is a jav builder
				local builderNum = player:GetClassAttributes():GetGambit(i+(player:GetClassAttributes():GetGambitCount() - self.maxGambitLength));
				if builderNum == 4 then
					builderNum = 1;
				end
				if self.builder[oppositeSequence+i] == builderNum and completeGambit ~= 0 then
					self.quickslotsBackColors[oppositeSequence+i]:SetBackColor(Turbine.UI.Color( 0,1,0 ));
				-- color incorrect spots red
				else
					self.quickslotsBackColors[oppositeSequence+i]:SetBackColor(Turbine.UI.Color( 1,0,0 ));
					completeGambit = 0; -- every spot after this will be red since 1 broken gambit makes the rest wrong
				end
			-- if you get to this point, the gambit spots are done
			else
				if completeGambit ~= 0 then
					completeGambit = 2;
				end
			end
		else
			-- for the builder slots that won't be needed, set their backgrounds to black
			if self.maxGambitLength - self.gambitLength >= i then
				self.quickslotsBackColors[i]:SetBackColor(Turbine.UI.Color( 0,0,0 ));
			-- for the ending gambits that should be used but aren't yet, set their backgrounds to white
			elseif i - (self.maxGambitLength - self.gambitLength) > player:GetClassAttributes():GetGambitCount() and completeGambit ~= 0 then
				self.quickslotsBackColors[i]:SetBackColor(Turbine.UI.Color( 1,1,1 ));
			else
				-- set backgrounds to correct gambits to green and incorrect ones to red
				-- gets the builder at position and changes it to 3 if it is a jav builder
				local builderNum = player:GetClassAttributes():GetGambit(i - (self.maxGambitLength - self.gambitLength));
				if builderNum == 4 then
					builderNum = 1;
				end
				if self.builder[i] == builderNum and completeGambit ~= 0 then
					self.quickslotsBackColors[i]:SetBackColor(Turbine.UI.Color( 0,1,0 ));
					if i == self.maxGambitLength then -- if this is true then the end of the gambits has been reached, set it to 2 if no problems came up
						completeGambit = 2;
					end
				else -- the gambit is wrong, set it to red
					self.quickslotsBackColors[i]:SetBackColor(Turbine.UI.Color( 1,0,0 ));
					completeGambit = 0; -- every spot after this will be red since 1 broken gambit makes the rest wrong
				end
			end
		end		
	end

	-- sets the position and data for the default gambit skill
	self.quickslots[self.maxGambitLength + 1]:SetShortcut( Turbine.UI.Lotro.Shortcut(Turbine.UI.Lotro.ShortcutType.Undefined, "") );
	self.quickslots[self.maxGambitLength + 1]:SetShortcut( Turbine.UI.Lotro.Shortcut( Turbine.UI.Lotro.ShortcutType.Skill, SDRPlugins.WardenEase2.Skills.Default.Data ) );
	self.quickslots[self.maxGambitLength + 1]:SetPosition( self.maxGambitLength * (self.quickslotSize + self.spacing), (rHeightLoc - 0) * multiplier );
	self.quickslots[self.maxGambitLength + 1]:SetVisible( true );
	if completeGambit == 2 then -- if completeGambit is 2, the gambit is complete, set the gambit color to green
		self.quickslotsBackColors[self.maxGambitLength+1]:SetBackColor(Turbine.UI.Color(0,1,0));
	elseif completeGambit == 1 then -- if completeGambit is 1, the gambit is not complete, but not wrong yet, set the gambit color to white
		self.quickslotsBackColors[self.maxGambitLength+1]:SetBackColor(Turbine.UI.Color(1,1,1));
	else self.quickslotsBackColors[self.maxGambitLength+1]:SetBackColor(Turbine.UI.Color(1,0,0)); -- else the gambit is wrong, set the gambit color to red
	end
	if wardenGambitChangedHandler == nil then
		wardenGambitChangedHandler = function(sender, args)
			self:RefreshLayout(); -- corrects the tooltips for the gambit
		end
		AddCallback(Turbine.Gameplay.Attributes.WardenAttributes, "GambitChanged", wardenGambitChangedHandler);
	end

	-- sets recovery skills
	if self.recoverySkills[SDRPlugins.WardenEase2.Recovery] and self.nRecoverySkills > 0  then
		self.quickslots[self.maxGambitLength + 2]:SetShortcut( Turbine.UI.Lotro.Shortcut( Turbine.UI.Lotro.ShortcutType.Skill, SDRPlugins.WardenEase2.Skills.Recovery.Data ) );
		self.quickslots[self.maxGambitLength + 2]:SetPosition( (self.maxGambitLength + modifier) * (self.quickslotSize + self.spacing) + self.extraSpacing, (rHeightLoc - 0) * multiplier );
		self.quickslots[self.maxGambitLength + 2]:SetVisible( true );
	end
	if self.recoverySkills[SDRPlugins.WardenEase2.QuickRecovery] and self.nRecoverySkills > 0  then
		local r = 3; -- normally the Quick Recovery skill goes into the maxGambitLength + 3 quickslot spot (since it goes after the default and recovery skills)
		local r2 = 1; -- normally the Quick Recovery skill goes into the maxGambitLength + 1 + modifier position
		-- if for some reason Recovery skill hasn't been taught, then Quick Recovery needs to go into it's place
		if self.nRecoverySkills == 1 then
			r = 2;
			r2 = 0;
		end
		self.quickslots[self.maxGambitLength + r]:SetShortcut( Turbine.UI.Lotro.Shortcut( Turbine.UI.Lotro.ShortcutType.Skill, SDRPlugins.WardenEase2.Skills.QuickRecovery.Data ) );
		self.quickslots[self.maxGambitLength + r]:SetPosition( (self.maxGambitLength + r2 + modifier) * (self.quickslotSize + self.spacing) + self.extraSpacing, (rHeightLoc - 0) * multiplier);
		self.quickslots[self.maxGambitLength + r]:SetVisible( true );
	end

	-- sets fast building skills
	if self.nFastBuildSkills > 0 then
		local j, l;

		-- sets the position of the fast building skills based on certain factors
		-- the number of skills across is always 3 (hence that 3)
		l = width / 2 - ((3 * (self.quickslotSize + self.spacing) - self.spacing) / 2);

		-- sets fast building skills
		for i = 1, self.nFastBuildSkills, 1 do
			j = i + self.maxGambitLength + 1 + self.nRecoverySkills; -- start at j to not override other skills' quickslots
			self.quickslots[j]:SetShortcut( Turbine.UI.Lotro.Shortcut(Turbine.UI.Lotro.ShortcutType.Undefined, "") );
			self.quickslots[j]:SetShortcut( Turbine.UI.Lotro.Shortcut( Turbine.UI.Lotro.ShortcutType.Skill, SDRPlugins.WardenEase2.Skills[self.settings.fastBuilderLocs[i]].Data ) );
			self.quickslots[j].index = i;
			-- the first 3 fast builders go on the top row
			if i < 4 then
				self.quickslots[j]:SetPosition( l + (i - 1) * (self.quickslotSize + self.spacing), (rHeightLoc - (self.quickslotSize + self.spacing)) * multiplier );
			-- the second 3 fast builders go in the middle row
			elseif i < 7 then
				self.quickslots[j]:SetPosition( l + (i - 4) * (self.quickslotSize + self.spacing), (rHeightLoc - ((self.quickslotSize + self.spacing)*2)) * multiplier );
			-- the last 3 fast builders go on the last row
			else
				self.quickslots[j]:SetPosition( l + (i - 7) * (self.quickslotSize + self.spacing), (rHeightLoc - ((self.quickslotSize + self.spacing)*3)) * multiplier );
			end
			self.quickslots[j]:SetVisible( true );

			-- create menu to allow user to change location of fast builders
			self.quickslots[j].MouseClick = function(sender, args)
				if args.Button == Turbine.UI.MouseButton.Right then
					local contextMenu = Turbine.UI.ContextMenu()
					local contextMenuItems = contextMenu:GetItems();
	
					-- create the "set" fast builder options
					for fastBuilder = 1, #self.settings.fastBuilderLocs, 1 do
						local subMenu = Turbine.UI.MenuItem(SDRPlugins.WardenEase2.SetSlotTo..SDRPlugins.WardenEase2[self.settings.fastBuilderLocs[fastBuilder]]);
						subMenu.FastBuilder = self.settings.fastBuilderLocs[fastBuilder];
						subMenu.Click = function( innerSender, args )
							local menuText = innerSender.FastBuilder;
							local oldLoc = 0;
							for oldLocFinder = 1, #self.settings.fastBuilderLocs, 1 do
								if self.settings.fastBuilderLocs[oldLocFinder] == menuText then
									oldLoc = oldLocFinder;
								end
							end
							local holder = self.settings.fastBuilderLocs[sender.index];
							self.settings.fastBuilderLocs[sender.index] = menuText;
							self.settings.fastBuilderLocs[oldLoc] = holder;
							self:RefreshQuickslots();
							self:SaveSettings();
						end
						contextMenuItems:Add( subMenu );
					end

					-- shows the context menu when the down arrow is clicked
					contextMenu:ShowMenu();


				end
			end
		end
	end

	-- sets the menu button position based on certain factors
	self.menuButton:SetPosition( ( width - self.menuButton:GetWidth() ) / 2, (((height - self.menuButton:GetHeight()) * self.settings.reversed) - ((self.quickslotSize + self.spacing) * self.slotsHigh + self.extraSpacing)) * multiplier);
	self.allowedChange = self.allowedChange - 1; -- resets allowed changed to prevent quickslots from being changed
end

-- determines whether a gambit is a legal and available name
-- gambitName is the name of the gambit - must be at least three characters in length!
-- returns the skillInfo if available, nil otherwise
function WEController:IsGambitAvailable(gambitName)
	if gambitName == nil or gambitName:len() < 3 then
		return nil;
	end
	gambitName = gambitName:lower();
	local gambits = player:GetClassAttributes():GetTrainedGambits();
	
	for i = 1, gambits:GetCount() do
		local gambitNameI = gambits:GetItem(i):GetSkillInfo():GetName():lower();
		if gambitNameI:find(gambitName) ~= nil then
			return gambits:GetItem(i):GetSkillInfo();
		end
	end
	return nil;
end